@@ -1,12 +1,73 @@
-import os
-import numpy as np
-import time, math
+import scanpy as sc
 import pandas as pd
+import numpy as np
+import math
 
 from scipy.sparse import csr_matrix, find
+import matplotlib.pyplot as plt
+import time
 from scipy.stats import norm
+
 from joblib import Parallel, delayed
 
+def which(x):
+    n = len(x)
+    m = sum(x)
+    y = np.zeros(m)
+    j = -1
+    for i in range(0, n):
+        if x[i]:
+            j = j+1
+            y[j] = i
+    y = y.astype(int)
+    return y
+
+def sparsetoid(sp_mtx, x = 0, y = 0):
+    (id_x, id_y) = sp_mtx.nonzero()
+    if len(id_x)*len(id_y) == 0:
+        v = np.array([])
+    else:
+        v = np.asarray(sp_mtx[id_x, id_y])[0]
+    d = {'x_id': id_x+x, 'y_id': id_y+y, 'value': v}
+    df = pd.DataFrame(data=d)
+    return df  
+
+def id_concat(pair):
+    frames = [pair[0], pair[1]]
+    result = pd.concat(frames)
+    return result
+
+def idtosparse(df, G1=None, G2=None):
+    if G1 is None:
+        G1 = max(df.x_id) + 1
+    if G2 is None:
+        G2 = max(df.y_id) + 1
+    sp_mtx = csr_matrix((df.value, (df.x_id, df.y_id)), shape=(G1, G2))
+    return sp_mtx
+
+def csntoflat(csn):
+    if type(csn) is list:
+        n2 = len(csn)
+        n1 = csn[0].shape[0]
+        csn_flat = np.zeros((int(n1*(n1-1)/2), n2))
+        k = 0
+        for i in range(0, n1-1):
+            for j in range(i+1, n1):
+                csn_flat[k, :] = np.asarray([item[i,j] for item in csn_loc_mat])
+                k = k + 1
+    else: 
+        (n1, n11, n2) = csn.shape
+        if n1 != n11: 
+            print('dimension not match!')
+            return
+        csn_flat = np.zeros((int(n1*(n1-1)/2), n2))
+        k = 0
+        for i in range(0, n1-1):
+            for j in range(i+1, n1):
+                csn_flat[k, :] = csn[i, j, :]
+                k = k + 1
+    return csn_flat
+
 def csn(data_full, g_mtx = None, wd_q = 0.1, dev = True, md = 1, iteration = False, fuzzy = False, ncore = 4):
     (n1, n2) = data_full.shape
     eps = np.finfo(float).eps
@@ -23,20 +84,33 @@ def csn(data_full, g_mtx = None, wd_q = 0.1, dev = True, md = 1, iteration = Fal
     print(*[L , 'pairs need calculation'])
     csn_mat = np.zeros((L, n2))
     
+    def valuetosparse(value, I, J, G1 = None, G2 = None):
+        if G1 is None:
+            G1 = max(df.x_id) + 1
+        if G2 is None:
+            G2 = max(df.y_id) + 1
+        I = I[value != 0]
+        J = J[value != 0]
+        value = value[value != 0]
+        sp_mtx = csr_matrix((value, (I, J)), shape = (G1, G2))
+        return sp_mtx
+    
     if dev:
         if fuzzy:
             selected_gene = np.unique([I, J])
             grid_gene = np.zeros((n1, n2))
             gene_first_zero = np.zeros(n1)
-            gene_mid = np.zeros((n1, 20))
             
             for s in selected_gene:
                 gene_temp = data_full[s, :]
                 max_gene = max(gene_temp[gene_temp!=0])
                 min_gene = min(gene_temp[gene_temp!=0])
                 range_temp = max_gene - min_gene
-                gene_cut = np.arange(min_gene, max_gene, range_temp/20)
-                gene_mid[s,:] = gene_cut[0:20] + range_temp/40
+                if range_temp == 0:
+                    gene_cut = np.array([min_gene])
+                else:
+                    gene_cut = np.arange(min_gene, max_gene, range_temp/20)
+                    
                 if sum(gene_temp == 0) > 1:
                     gene_cut = np.insert(gene_cut, 0, 0)
                     gene_first_zero[s] = 1
@@ -71,7 +145,7 @@ def csn(data_full, g_mtx = None, wd_q = 0.1, dev = True, md = 1, iteration = Fal
                 csn_temp = np.zeros(n2)
                 for t in range(0, len(cell_id)):
                     k = cell_id[t]
-                    B = np.zeros(2, n2)
+                    B = np.zeros((2, n2))
                     for l in range(0, n2):
                         B[:, l] = (data[:,l] <= upper[:, k]) & (data[:, l] >= lower[:, k]) & (data[:, k] > 0)
                     a = B.sum(axis = 1)
@@ -81,6 +155,7 @@ def csn(data_full, g_mtx = None, wd_q = 0.1, dev = True, md = 1, iteration = Fal
                 
                 return csn_temp
             csn_arr_temp = np.asarray(Parallel(n_jobs = ncore)(delayed(inner_fuzzy_fun)(m) for m in range(0, L)))
+            
             csn = [valuetosparse(v, I, J, n1, n1) for v in list(csn_arr_temp.T)]
         else:
             def inner_fun(m):
@@ -102,6 +177,7 @@ def csn(data_full, g_mtx = None, wd_q = 0.1, dev = True, md = 1, iteration = Fal
                         csn_temp[k] = temp[0, 1]
                 return csn_temp
             csn_arr_temp = np.asarray(Parallel(n_jobs = ncore)(delayed(inner_fun)(m) for m in range(0, L)))
+            
             csn = [valuetosparse(v, I, J, n1, n1) for v in list(csn_arr_temp.T)]
             
     else: 
@@ -312,15 +388,17 @@ def csn_rec(data1, data2, g_mtx = None, wd_q = 0.1, dev = True, md = 1, iteratio
             selected_gene = [np.unique(I), np.unique(J)]
             grid_gene = [np.zeros((G1, N)), np.zeros((G2, N))]
             gene_first_zero = [np.zeros(G1), np.zeros(G2)]
-            gene_mid = [np.zeros((G1, 20)), np.zeros((G2, 20))]
+            
             for k in range(0, 1):
                 for s in selected_gene[k]:
                     gene_temp = data[k][s, :]
                     max_gene = max(gene_temp[gene_temp!=0])
                     min_gene = min(gene_temp[gene_temp!=0])
                     range_temp = max_gene - min_gene
-                    gene_cut = np.arange(min_gene, max_gene, range_temp/20)
-                    gene_mid[k][s,:] = gene_cut[0:20] + range_temp/40
+                    if range_temp == 0:
+                        gene_cut = np.array([min_gene])
+                    else:
+                        gene_cut = np.arange(min_gene, max_gene, range_temp/20)
                     if sum(gene_temp == 0) > 1:
                         gene_cut = np.insert(gene_cut, 0, 0)
                         gene_first_zero[k][s] = 1
@@ -355,7 +433,7 @@ def csn_rec(data1, data2, g_mtx = None, wd_q = 0.1, dev = True, md = 1, iteratio
                 csn_temp = np.zeros(N)
                 for t in range(0, len(cell_id)):
                     k = cell_id[t]
-                    B = np.zeros(2, N)
+                    B = np.zeros((2, N))
                     for l in range(0, N):
                         B[:, l] = (data[:,l] <= upper[:, k]) & (data[:, l] >= lower[:, k]) & (data[:, k] > 0)
                     a = B.sum(axis = 1)
@@ -390,6 +468,17 @@ def csn_rec(data1, data2, g_mtx = None, wd_q = 0.1, dev = True, md = 1, iteratio
                 return csn_temp
             csn_arr_temp = np.asarray(Parallel(n_jobs = ncore)(delayed(inner_fun)(m) for m in range(0, L)))
 
+        def valuetosparse(value, I, J, G1 = None, G2 = None):
+            if G1 is None:
+                G1 = max(df.x_id) + 1
+            if G2 is None:
+                G2 = max(df.y_id) + 1
+            I = I[value != 0]
+            J = J[value != 0]
+            value = value[value != 0]
+            sp_mtx = csr_matrix((value, (I, J)), shape = (G1, G2))
+            return sp_mtx
+        
         csn = [valuetosparse(v, I, J, G1, G2) for v in list(csn_arr_temp.T)]
     else: 
         print('Please use csn directly')
@@ -416,13 +505,13 @@ def csn_block(data, M = 100, g_mtx = None, wd_q = 0.1, dev = True, md = 1, itera
         data_i = data[group_n == i, :]
         for j in range(i, n):
             if i == j:
-                g_mtx_temp = g_mtx[group_n == i, group_n == i]
+                g_mtx_temp = g_mtx[group_n == i,:][:, group_n == i]
                 csn_temp = csn(data_i, g_mtx_temp, wd_q, dev, md, iteration, fuzzy, ncore)
                 csn_id_temp = [sparsetoid(item, i*M, j*M) for item in csn_temp]
                 csn_mtx_id = [id_concat(pair) for pair in zip(csn_mtx_id, csn_id_temp)]
             else:
                 data_j = data[group_n == j,:]
-                g_mtx_temp = g_mtx[group_n == i, group_n == j]
+                g_mtx_temp = g_mtx[group_n == i,:][:, group_n == j]
                 csn_temp = csn_rec(data_i, data_j, g_mtx_temp, wd_q, dev, md, iteration, fuzzy, ncore)
                 csn_id_temp = [sparsetoid(item, i*M, j*M) for item in csn_temp]
                 csn_mtx_id = [id_concat(pair) for pair in zip(csn_mtx_id, csn_id_temp)]
@@ -572,3 +661,54 @@ def csn_block_loc(data, knn_index, M = 100, wd_q = 0.1, dev = True, md = 1, iter
     csn_mtx = [idtosparse(item, G, G) for item in csn_mtx_id]
     return csn_mtx
    
+def create_D_mat(csn_mat, ncore = 4):
+    if type(csn_mat) is list:
+        T = len(csn_mat)
+    else:
+        T = csn_mat.shape[2]
+    D_mat = np.zeros((T, T))
+    s_mtx = np.ones((T, T)) - np.tri(T)
+    (I, J, S) = find(s_mtx)
+    L = len(I)
+    if type(csn_mat) is list:
+        def inner_fun(m):
+            i = I[m]
+            j = J[m]
+            a = sum(abs(csn_mat[i]-csn_mat[j])).max()
+            return a
+        A = np.asarray(Parallel(n_jobs = ncore)(delayed(inner_fun)(m) for m in range(0, L)))
+        D_mat[I, J] = A
+        D_mat[J, I] = A
+    else:
+        def inner_fun(m):
+            i = I[m]
+            j = J[m]
+            a = max(sum(abs(csn_mat[:, :, i] - csn_mat[:, :, j])))
+            return a
+        A = np.asarray(Parallel(n_jobs = ncore)(delayed(inner_fun)(m) for m in range(0, L)))
+        D_mat[I, J] = A
+        D_mat[J, I] = A
+    return D_mat
+    
+def distance_test(D_mat, change, num_perm = 1000):
+    stat_rec = np.zeros(num_perm)
+    n = D_mat.shape[0]
+    Z = np.zeros((n, 2))
+    Z[0:change, 0] = 1
+    Z[change:, 1] = -1
+    
+    counts = sum(abs(Z))
+    counts = np.reshape(counts, (1, 2))
+    count_mat = counts.T@counts
+    for i in range(0, num_perm):
+        if i == 0:
+            ind = np.asarray(list(range(0, n)))
+        else:
+            ind = np.random.permutation(n)
+        Dp = D_mat[ind, :][:, ind]
+        test_stat = Z.T@Dp@Z / (count_mat + pow(10, -6))
+        stat_rec[i] = sum(sum(test_stat))
+    
+    pval = 1 - (stat_rec > stat_rec[0] + pow(10, -6)).mean()
+    return pval
+
